﻿using Silk.NET.Maths;
using Silk.NET.OpenGL;

namespace Nebulix.Rendering
{
	public sealed class Mesh
	{
		public Vector3D<float>[] Vertices { get => vertices; set { vertices = value; regenerate = true; } }
		public nuint[] Indices { get => indices; set { indices = value; regenerate = true; } }
		
		private uint vao = 0, vbo = 0, ebo = 0;
		private bool regenerate = true;

		private Vector3D<float>[] vertices = [];
		private nuint[] indices = [];
		private Vector3D<float>[] normals = [];

		public void Clear()
		{
			vertices = [];
			indices = [];
		}

		public void CalculateNormals()
		{
			normals = new Vector3D<float>[vertices.Length];
			for (int j = 0; j < vertices.Length; j++)
			{
				normals[j] = Vector3D<float>.Zero;
			}

			return;
			for (int i = 0; i < indices.Length; i += 3)
			{
				Vector3D<float> a = vertices[indices[i] - 1];
				Vector3D<float> b = vertices[indices[i + 1] - 1];
				Vector3D<float> c = vertices[indices[i + 2] - 1];

				Vector3D<float> normal = Vector3D.Cross(b-a, c-a);
				normal = Vector3D.Normalize(normal);
				normals[indices[i] - 1] = normal;
				normals[indices[i + 1] - 1] = normal;
				normals[indices[i + 2] - 1] = normal;
			}
		}
		
		// getting called by "Engine" which currently is in another assembly, meaning I probably need to make this public
		// needs to be change for the real engine
		public void Render(GL gl)
		{
			if (regenerate) Generate(gl);

			gl.BindVertexArray(vao);
			gl.DrawElements(PrimitiveType.Triangles, (uint)vertices.Length * 3, DrawElementsType.UnsignedInt, 0);
		}

		private unsafe void Generate(GL gl)
		{
			regenerate = false;

			if(vao == 0)
				vao = gl.CreateVertexArray();
			if(vbo == 0)
				vbo = gl.GenBuffer();
			if(ebo == 0) 
				ebo = gl.GenBuffer();

			gl.BindVertexArray(vao);

			List<float> meshData = new(vertices.Length * 3 + normals.Length * 3);
			meshData.AddRange(vertices.ExtractComponents());
			//meshData.AddRange(normals.ExtractComponents());
			ReadOnlySpan<float> data = new(meshData.ToArray());
			gl.BindBuffer(BufferTargetARB.ArrayBuffer, vbo);
			gl.BufferData(BufferTargetARB.ArrayBuffer, (nuint)(data.Length * sizeof(float)), data, BufferUsageARB.StaticDraw);

			ReadOnlySpan<nuint> indicesData = new(indices);
			gl.BindBuffer(BufferTargetARB.ElementArrayBuffer, ebo);
			gl.BufferData(BufferTargetARB.ElementArrayBuffer, (nuint)(indicesData.Length * sizeof(nuint)), indicesData, BufferUsageARB.StaticDraw);

			// vertices
			gl.EnableVertexAttribArray(0);
			gl.VertexAttribPointer(0, 3, VertexAttribPointerType.Float, false, 3 * sizeof(float), null);
			// normals
			//gl.EnableVertexAttribArray(1);
			//gl.VertexAttribPointer(1, 3, VertexAttribPointerType.Float, false, 6 * sizeof(float), (void*)(3 * sizeof(float)));
		}
	}
}
